"""Track statistics generation module.

Author:
    Paulo Sanchez (@erlete)
"""


from ..api.track import TrackAPI
from ..core.vector import Rotator3D, Vector3D


class TrackStatistics:

    def __init__(
        self,
        track: TrackAPI,
        timestep: int | float
    ) -> None:
        """Initialize a TrackStatistics instance.

        Args:
            track (TrackAPI): statistics track.
            timestep (int | float): statistics timestep.
        """
        self.track = track
        self.timestep = timestep
        self._waypoints = track.waypoints

        # Automatically generated attributes:
        self._is_completed = False
        self._distance_to_end = 0.0
        self._data = [(track.track.start, Rotator3D(), 0.0)]  # Initial data.

    @property
    def track(self) -> TrackAPI:
        """Get track.

        Returns:
            TrackAPI: track.
        """
        return self._track

    @track.setter
    def track(self, value: TrackAPI) -> None:
        """Set track.

        Args:
            value (TrackAPI): track.
        """
        if not isinstance(value, TrackAPI):
            raise TypeError(
                "expected type TrackAPI for"
                + f" {self.__class__.__name__}.track but got"
                + f" {type(value).__name__} instead"
            )

        self._track = value

    @property
    def timestep(self) -> int | float:
        """Get statistics timestep.

        Returns:
            int | float: statistics timestep.
        """
        return self._timestep

    @timestep.setter
    def timestep(self, value: int | float) -> None:
        """Set statistics timestep.

        Args:
            value (int | float): statistics timestep.
        """
        if not isinstance(value, (int, float)):
            raise TypeError(
                "expected type int | float for"
                + f" {self.__class__.__name__}.timestep but got"
                + f" {type(value).__name__} instead"
            )

        self._timestep = value

    @property
    def waypoints(self) -> list[Vector3D]:
        """Get track waypoints.

        Returns:
            list[Vector3D]: track waypoints.
        """
        return self._waypoints

    @property
    def is_completed(self) -> bool:
        """Get track completion status.

        Returns:
            bool: track completion status.
        """
        return self._is_completed

    @is_completed.setter
    def is_completed(self, value: bool) -> None:
        """Set track completion status.

        Args:
            value (bool): track completion status.
        """
        if not isinstance(value, bool):
            raise TypeError(
                "expected type bool for"
                + f" {self.__class__.__name__}.is_completed but got"
                + f" {type(value).__name__} instead"
            )

        self._is_completed = value

    @property
    def distance_to_end(self) -> float:
        """Get drone distance to track end.

        Returns:
            float: drone distance to track end.
        """
        return self._distance_to_end

    @distance_to_end.setter
    def distance_to_end(self, value: int | float) -> None:
        """Set drone distance to track end.

        Args:
            value (int | float): drone distance to track end.
        """
        if not isinstance(value, (int, float)):
            raise TypeError(
                "expected type int | float for"
                + f" {self.__class__.__name__}.distance_to_end but got"
                + f" {type(value).__name__} instead"
            )

        self._distance_to_end = float(value)

    @property
    def data(self) -> list[tuple[Vector3D, Rotator3D, int | float]]:
        """Get drone position, rotation and speed data at each timestep.

        Returns:
            list[tuple[Vector3D, Rotator3D, int | float]]: position,
                rotation and speed of the drone at each timestep.
        """
        return self._data

    @property
    def positions(self) -> list[Vector3D]:
        """Get drone positions at each timestep.

        Returns:
            list[Vector3D]: drone positions at each timestep.
        """
        return [data[0] for data in self._data]

    @property
    def rotations(self) -> list[Rotator3D]:
        """Get drone rotations at each timestep.

        Returns:
            list[Rotator3D]: drone rotations at each timestep.
        """
        return [data[1] for data in self._data]

    @property
    def speeds(self) -> list[int | float]:
        """Get drone speeds at each timestep.

        Returns:
            list[int | float]: drone speeds at each timestep.
        """
        return [data[2] for data in self._data]

    def add_data(
        self,
        position: Vector3D,
        rotation: Rotator3D,
        speed: int | float
    ) -> None:
        """Add drone position, rotation and speed data.

        Args:
            position (Vector3D): drone position.
            rotation (Rotator3D): drone rotation.
            speed (int | float): drone speed.
        """
        if not isinstance(position, Vector3D):
            raise TypeError(
                "expected type Vector3D for"
                + f" {self.__class__.__name__}.add_data but got"
                + f" {type(position).__name__} instead"
            )

        if not isinstance(rotation, Rotator3D):
            raise TypeError(
                "expected type Rotator3D for"
                + f" {self.__class__.__name__}.add_data but got"
                + f" {type(rotation).__name__} instead"
            )

        if not isinstance(speed, (int, float)):
            raise TypeError(
                "expected type int | float for"
                + f" {self.__class__.__name__}.add_data but got"
                + f" {type(speed).__name__} instead"
            )

        self._data.append((position, rotation, speed))



"""Environment management module.

This module allows 3D environment generation from templates, data structuring
and display in an easy way.

Modules:
    reader: track template reader module.
    track: track structuring and display module.

Author:
    Paulo Sanchez (@erlete)
"""



"""Track utilities module.

Author:
    Paulo Sanchez (@erlete)
"""


from ..core.gradient import ColorGradient
from ..core.vector import Vector3D
from ..geometry.ring import Ring


class Track:
    """Track representation class.

    This class is used to represent a track composed of a start point, an end
    point and a sequence of rings.

    Attributes:
        start (Vector3D): track start.
        end (Vector3D): track end.
        rings (list[Ring]): track rings.
        waypoints (list[Vector3D]): track waypoints.
    """

    def __init__(
        self,
        start: Vector3D,
        end: Vector3D,
        rings: list[Ring]
    ) -> None:
        """Initialize a Track instance.

        Args:
            start (Vector3D): track start.
            end (Vector3D): track end.
            rings (list[Ring]): track rings.
        """
        self.start = start
        self.end = end
        self.rings = rings

    @property
    def start(self) -> Vector3D:
        """Get track start.

        Returns:
            Vector3D: track start.
        """
        return self._start

    @start.setter
    def start(self, value: Vector3D) -> None:
        """Set track start.

        Args:
            value (Vector3D): track start.
        """
        if not isinstance(value, Vector3D):
            raise TypeError(
                "expected type Vector3D for"
                + f" {self.__class__.__name__}.start but got"
                + f" {type(value).__name__} instead"
            )

        self._start = value

    @property
    def end(self) -> Vector3D:
        """Get track end.

        Returns:
            Vector3D: track end.
        """
        return self._end

    @end.setter
    def end(self, value: Vector3D) -> None:
        """Set track end.

        Args:
            value (Vector3D): track end.
        """
        if not isinstance(value, Vector3D):
            raise TypeError(
                "expected type Vector3D for"
                + f" {self.__class__.__name__}.end but got"
                + f" {type(value).__name__} instead"
            )

        self._end = value

    @property
    def rings(self) -> list[Ring]:
        """Get track rings.

        Returns:
            list[Ring]: track rings.
        """
        return self._rings

    @rings.setter
    def rings(self, value: list[Ring]) -> None:
        """Set track rings.

        Args:
            value (list[Ring]): track rings.
        """
        if not isinstance(value, list):
            raise TypeError(
                "expected type list[Ring] for"
                + f" {self.__class__.__name__}.rings but got"
                + f" {type(value).__name__} instead"
            )

        for element in value:
            if not isinstance(element, Ring):
                raise TypeError(
                    f"expected Ring, got {type(element).__name__}"
                    + f"on element {element} instead"
                )

        self._rings = value

    @property
    def waypoints(self) -> list[Vector3D]:
        """Get track waypoints.

        Returns:
            list[Vector3D]: track waypoints.
        """
        return [self.start, *[ring.position for ring in self._rings], self.end]

    @staticmethod
    def ax_auto_fit(ax, offset: int = 1, *waypoints: Vector3D) -> None:
        """Set axis limits automatically.

        Args:
            ax (Axes3D): ax to set limits for.
            offset (int, optional): additional distance margin. Defaults to 1.
        """
        x = [checkpoint.x for checkpoint in waypoints]
        y = [checkpoint.y for checkpoint in waypoints]
        z = [checkpoint.z for checkpoint in waypoints]

        centers = (
            (max(x) + min(x)) / 2,
            (max(y) + min(y)) / 2,
            (max(z) + min(z)) / 2
        )

        max_distance = max(
            abs(max(x) - min(x)) / 2,
            abs(max(y) - min(y)) / 2,
            abs(max(z) - min(z)) / 2
        )

        ax.set_xlim3d(
            centers[0] - max_distance - offset,
            centers[0] + max_distance + offset
        )
        ax.set_ylim3d(
            centers[1] - max_distance - offset,
            centers[1] + max_distance + offset
        )
        ax.set_zlim3d(
            centers[2] - max_distance - offset,
            centers[2] + max_distance + offset
        )

    def plot(self, ax, **kwargs) -> None:
        """Plot track.

        Args:
            ax (Axes3D): 3D axis.
            **kwargs: plot arguments.
        """
        self.start.plot(ax, "+", color="darkred", **kwargs)
        self.end.plot(ax, "P", color="darkgreen", **kwargs)

        # Color gradient for rings:
        gradient = ColorGradient("#ff0000", "#0000ff", len(self.rings))
        steps = [ColorGradient.rgb_to_hex(step) for step in gradient.steps]

        # Ring plotting:
        for color, ring in zip(steps, self.rings):
            ring.plot(
                ax,
                color=color,
                edgecolors=color,
                **kwargs
            )

        # Title, labels and legend:
        ax.set_title("Drone track view")
        ax.set_xlabel("X")
        ax.set_ylabel("Y")
        ax.set_zlabel("Z")
        ax.legend(
            [
                "Start",
                "End"
            ] + [
                f"Ring {i}"
                for i in range(1, len(self.rings) + 1)
            ]
        )

        # Axis planes background color:
        ax.xaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
        ax.yaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
        ax.zaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))

        # Grid and axis views:
        # ax.grid(False)  # TODO: Activate this after debugging
        # ax.set_axis_off()  # TODO: Activate this after debugging
        ax.set_facecolor((1.0, 1.0, 1.0, 0.0))

        # Projection, aspect and axis scale:
        ax.set_proj_type("ortho")
        ax.set_box_aspect((1, 1, 1))
        ax.set_autoscale_on(False)
        self.ax_auto_fit(
            ax,
            1,
            self.start,
            self.end,
            *[ring.position for ring in self.rings]
        )

        # View distance and angle:
        ax.view_init(azim=-135, elev=45)
        ax.dist = 10

    def __repr__(self) -> str:
        """Get short track representation.

        Returns:
            str: short track representation.
        """
        return f"<Track with {len(self._rings)} rings>"

    def __str__(self) -> str:
        """Get long track representation.

        Returns:
            str: long track representation.
        """
        return (
            f"Track from {self._start} to {self._end} with"
            + f" rings: {self._rings}"
        )

    def __len__(self) -> int:
        """Get track length.

        Returns:
            int: track length.
        """
        return len(self._rings) + 2  # Start and end points compensation.

    def __hash__(self) -> int:
        """Get track hash.

        Returns:
            int: track hash.
        """
        return hash(tuple([self._start, *self._rings, self._end]))



"""Core functionalities module.

This module contains all classes required to provide with the core
functionality, including vector operations and display, color gradient
generation, etc.

Modules:
    color: color utilities module.
    vector: vector utilities module.

Author:
    Paulo Sanchez (@erlete)
"""



"""API module.

This module is perhaps the most important and complex one, as it provides with
the Simulation API, which allows the user to control a drone along a set of
designed tracks. The API automatically handles track switching, so the
information it provides is always related to the current track. It also
provides with statistical measurement tools for keeping track on the simulation
progress.

Modules:
    drone: drone API extension.
    simulation: core simulation API.
    statistics: statistical measurement tools.
    track: track API extension.

Author:
    Paulo Sanchez (@erlete)
"""



"""Geometry module.

This module contains several classes that make it possible to display specific
surfaces in a 3D environment, such as track ring and drone models.

Modules:
    drone: drone model display module.
    ring: track ring display module.

Author:
    Paulo Sanchez (@erlete)
"""



"""Color grading generator tools container module.

Author:
    Paulo Sanchez (@erlete)
"""


class ColorGradient:
    """Color gradient representation class.

    This class is used to generate hexadecimal color gradient steps between two
    colors.

    Attributes:
        initial_color (str): gradient initial color.
        final_color (str): gradient final color.
        step_count (int): gradient step count.
        steps (list[list[int]]): gradient steps.
    """

    def __init__(
        self,
        initial_color: str,
        final_color: str,
        step_count: int
    ) -> None:
        """Initialize a ColorGradient instance.

        Args:
            initial_color (str): gradient initial color.
            final_color (str): gradient final color.
            step_count (int): gradient step count.
        """
        self.initial_color = initial_color
        self.final_color = final_color
        self.step_count = step_count

    @property
    def initial_color(self) -> str:
        """Get gradient initial color.

        Returns:
            str: gradient initial color.
        """
        return self._initial_color

    @initial_color.setter
    def initial_color(self, value: str) -> None:
        """Set gradient initial color.

        Args:
            value (str): gradient initial color.
        """
        if not isinstance(value, str):
            raise TypeError(
                "expected type str for"
                + f" {self.__class__.__name__}.initial_color but got"
                + f" {type(value).__name__} instead"
            )

        if not self.is_valid_hex(value):
            raise ValueError(
                f"invalid hex color {value} for"
                + f" {self.__class__.__name__}.initial_color"
            )

        self._initial_color = value

    @property
    def final_color(self) -> str:
        """Get gradient final color.

        Returns:
            str: gradient final color.
        """
        return self._final_color

    @final_color.setter
    def final_color(self, value: str) -> None:
        """Set gradient final color.

        Args:
            value (str): gradient final color.
        """
        if not isinstance(value, str):
            raise TypeError(
                "expected type str for"
                + f" {self.__class__.__name__}.final_color but got"
                + f" {type(value).__name__} instead"
            )

        if not self.is_valid_hex(value):
            raise ValueError(
                f"invalid hex color {value} for"
                + f" {self.__class__.__name__}.final_color"
            )

        self._final_color = value

    @property
    def step_count(self) -> int:
        """Get gradient step count.

        Returns:
            int: gradient step count.
        """
        return self._step_count

    @step_count.setter
    def step_count(self, value: int) -> None:
        """Set gradient step count.

        Args:
            value (int): gradient step count.
        """
        if not isinstance(value, int):
            raise TypeError(
                "expected type int for"
                + f" {self.__class__.__name__}.step_count but got"
                + f" {type(value).__name__} instead"
            )

        self._step_count = value

    @property
    def steps(self) -> list[list[int]]:
        """Get gradient steps.

        Returns:
            list[list[int]]: gradient steps.
        """
        return self._compute_steps() if self._step_count > 2 else [
            self.hex_to_rgb(self._initial_color),
            self.hex_to_rgb(self._final_color)
        ]

    @staticmethod
    def is_valid_hex(hex_str: str) -> bool:
        """Check if a string is a valid hex color.

        Args:
            hex_str (str): string to check.

        Returns:
            bool: True if the string is a valid hex color, False otherwise.
        """
        return (
            isinstance(hex_str, str)
            and len(hex_str) == 7
            and hex_str[0] == "#"
            and hex_str[1:].isalnum()
        )

    @staticmethod
    def rgb_to_hex(rgb: list[int]) -> str:
        """Convert an RGB color sequence hex string.

        Args:
            rgb (list[int]): RGB list.

        Returns:
            str: hex color string.
        """
        return f"#{rgb[0]:02x}{rgb[1]:02x}{rgb[2]:02x}"

    @staticmethod
    def hex_to_rgb(hex_str: str) -> list[int]:
        """Convert a hex color string to RGB.

        Args:
            hex_str (str): hex color string.

        Returns:
            list[int]: RGB list.
        """
        return [int(hex_str[i:i + 2], 16) for i in range(1, 6, 2)]

    def _compute_steps(self) -> list[list[int]]:
        """Compute gradient steps.

        Returns:
            list[list[int]]: gradient steps.
        """
        initial_rgb = self.hex_to_rgb(self.initial_color)
        final_rgb = self.hex_to_rgb(self.final_color)

        return [
            [
                int(
                    initial_rgb[i] + (final_rgb[i] - initial_rgb[i])
                    * step / self._step_count
                ) for i in range(3)
            ] for step in range(self._step_count)
        ]



"""Simple Drone Control (SDC) package.

This package contains several libraries that allow the user to control a drone
in a simulated 3D environment. The model is purely mathematical and does not
consider certain physical or dynamic aspects of the real world, since it is
meant to be used as a learning tool.

Modules:
    api: API modules for simulation, tracks and drones.
    core: core modules such as vector and color utilities.
    environment: environment management and display module.
    geometry: geometry generation and display module.

Author:
    Paulo Sanchez (@erlete)
"""



"""Drone API module.

Author:
    Paulo Sanchez (@erlete)
"""


from ..core.vector import Rotator3D, Vector3D
from ..geometry.drone import Drone


class DroneAPI(Drone):
    """Drone API class.

    This class represents a kinematic drone model, implementing the geometry
    of the drone and adding a speed attribute to it.

    Attributes:
        drone (Drone): drone.
        speed (float): drone speed in m/s.
        SPEED_RANGE (tuple[int, int]): allowed drone speed range in m/s.
    """

    SPEED_RANGE = (0, 20)  # [m/s]

    def __init__(
        self,
        position: Vector3D,
        rotation: Rotator3D,
        speed: int | float = 0
    ) -> None:
        """Initialize a DroneAPI instance.

        Args:
            position (Vector3D): drone position.
            rotation (Rotator3D): drone rotation.
            speed (int | float): drone speed in m/s.
        """
        super().__init__(position, rotation)
        self.speed = speed

    @property
    def speed(self) -> float:
        """Get drone speed.

        Returns:
            float: drone speed.
        """
        return self._speed

    @speed.setter
    def speed(self, value: int | float) -> None:
        """Set drone speed.

        Args:
            value (int | float): drone speed.
        """
        if not isinstance(value, (int, float)):
            raise TypeError(
                "expected type float for"
                + f" {self.__class__.__name__}.speed but got"
                + f" {type(value).__name__} instead"
            )

        self._speed = float(
            max(self.SPEED_RANGE[0], min(value, self.SPEED_RANGE[1]))
        )

    def __repr__(self) -> str:
        """Get short drone representation.

        Returns:
            str: short drone representation.
        """
        return f"<DroneAPI at {self._position}>"

    def __str__(self) -> str:
        """Get long drone representation.

        Returns:
            str: long drone representation.
        """
        return f"""DroneAPI(
    position={self._position},
    rotation={self._rotation},
    speed={self._speed}
)"""



"""Vector utilities container module.

This module contains all utilities related to vector generation and
manipulation operations.

Author:
    Paulo Sanchez (@erlete)
"""


from __future__ import annotations

import math

import numpy as np


class Vector3D:
    """3D vector representation class.

    This class represents a vector in the 3D space. It can also be used to
    represent coordinates in a cartesian coordinate system.

    Attributes:
        x (float): X component of the vector.
        y (float): Y component of the vector.
        z (float): Z component of the vector.
    """

    def __init__(
        self,
        x: int | float = 0,
        y: int | float = 0,
        z: int | float = 0
    ) -> None:
        """Initialize a Vector3D instance.

        Args:
            x (int | float): X component of the vector.
            y (int | float): Y component of the vector.
            z (int | float): Z component of the vector.
        """
        self.x = x
        self.y = y
        self.z = z

    @property
    def x(self) -> float:
        """Get X component of the vector.

        Returns:
            float: X component of the vector.
        """
        return self._x

    @x.setter
    def x(self, value: int | float) -> None:
        """Set X component of the vector.

        Args:
            value (int | float): X component of the vector.
        """
        if not isinstance(value, (int, float)):
            raise TypeError(
                "expected type int | float for"
                + f" {self.__class__.__name__}.x but got"
                + f" {type(value).__name__} instead"
            )

        self._x = float(value)

    @property
    def y(self) -> float:
        """Get Y component of the vector.

        Returns:
            float: Y component of the vector.
        """
        return self._y

    @y.setter
    def y(self, value: int | float) -> None:
        """Set Y component of the vector.

        Args:
            value (int | float): Y component of the vector.
        """
        if not isinstance(value, (int, float)):
            raise TypeError(
                "expected type int | float for"
                + f" {self.__class__.__name__}.y but got"
                + f" {type(value).__name__} instead"
            )

        self._y = float(value)

    @property
    def z(self) -> float:
        """Get Z component of the vector.

        Returns:
            float: Z component of the vector.
        """
        return self._z

    @z.setter
    def z(self, value: int | float) -> None:
        """Set Z component of the vector.

        Args:
            value (int | float): Z component of the vector.
        """
        if not isinstance(value, (int, float)):
            raise TypeError(
                "expected type int | float for"
                + f" {self.__class__.__name__}.z but got"
                + f" {type(value).__name__} instead"
            )

        self._z = float(value)

    def plot(self, ax, *args, **kwargs) -> None:
        """Plot the vector.

        Args:
            ax (Axes3D): axes to plot the vector on.
        """
        ax.plot(*[[c, c] for c in self], *args, **kwargs)

    def __add__(self, other: Vector3D) -> Vector3D:
        """Add two vectors.

        Args:
            other (Vector3D): vector to add.

        Returns:
            Vector3D: resulting vector.
        """
        return Vector3D(
            self.x + other.x,
            self.y + other.y,
            self.z + other.z
        )

    def __sub__(self, other: Vector3D) -> Vector3D:
        """Subtract two vectors.

        Args:
            other (Vector3D): vector to subtract.

        Returns:
            Vector3D: resulting vector.
        """
        return Vector3D(
            self.x - other.x,
            self.y - other.y,
            self.z - other.z
        )

    def __mul__(self, other: int | float) -> Vector3D:
        """Multiply a vector by a scalar.

        Args:
            other (int | float): scalar to multiply by.

        Returns:
            Vector3D: resulting vector.
        """
        if not isinstance(other, (int, float)):
            raise TypeError("only scalars are supported for multiplication")

        return Vector3D(
            self.x * other,
            self.y * other,
            self.z * other
        )

    def __rmul__(self, other: int | float) -> Vector3D:
        """Multiply a vector by a scalar.

        Args:
            other (int | float): scalar to multiply by.

        Returns:
            Vector3D: resulting vector.
        """
        if not isinstance(other, (int, float)):
            raise TypeError("only scalars are supported for multiplication")

        return Vector3D(
            self.x * other,
            self.y * other,
            self.z * other
        )

    def __truediv__(self, other: int | float) -> Vector3D:
        """Divide a vector by a scalar (float division).

        Args:
            other (int | float): scalar to divide by.

        Returns:
            Vector3D: resulting vector.
        """
        if not isinstance(other, (int, float)):
            raise TypeError("only scalars are supported for division")

        return Vector3D(
            self.x / other,
            self.y / other,
            self.z / other
        )

    def __floordiv__(self, other: int | float) -> Vector3D:
        """Divide a vector by a scalar (integer division).

        Args:
            other (int | float): scalar to divide by.

        Returns:
            Vector3D: resulting vector.
        """
        if not isinstance(other, (int, float)):
            raise TypeError("only scalars are supported for division")

        return Vector3D(
            self.x // other,
            self.y // other,
            self.z // other
        )

    def __mod__(self, other: int | float) -> Vector3D:
        """Get the remainder of a vector division by a scalar.

        Args:
            other (int | float): scalar to divide by.

        Returns:
            Vector3D: resulting vector.
        """
        if not isinstance(other, (int, float)):
            raise TypeError("only scalars are supported for modulo")

        return Vector3D(
            self.x % other,
            self.y % other,
            self.z % other
        )

    def __pow__(self, other: int | float) -> Vector3D:
        """Get the power of a vector.

        Args:
            other (int | float): scalar to raise to.

        Returns:
            Vector3D: resulting vector.
        """
        if not isinstance(other, (int, float)):
            raise TypeError("only scalars are supported for exponentiation")

        return Vector3D(
            self.x ** other,
            self.y ** other,
            self.z ** other
        )

    def __abs__(self) -> Vector3D:
        """Get the absolute value of a vector.

        Returns:
            Vector3D: resulting vector.
        """
        return Vector3D(
            abs(self.x),
            abs(self.y),
            abs(self.z)
        )

    def __neg__(self) -> Vector3D:
        """Get the negation of a vector.

        Returns:
            Vector3D: resulting vector.
        """
        return Vector3D(
            -self.x,
            -self.y,
            -self.z
        )

    def __pos__(self) -> Vector3D:
        """Get the positive value of a vector.

        Returns:
            Vector3D: resulting vector.
        """
        return Vector3D(
            +self.x,
            +self.y,
            +self.z
        )

    def __round__(self, n: int = 0) -> Vector3D:
        """Round a vector.

        Args:
            n (int, optional): number of decimals to round to. Defaults to 0.

        Returns:
            Vector3D: resulting vector.
        """
        return Vector3D(
            round(self.x, n),
            round(self.y, n),
            round(self.z, n)
        )

    def __floor__(self) -> Vector3D:
        """Get the floor of a vector.

        Returns:
            Vector3D: resulting vector.
        """
        return Vector3D(
            math.floor(self.x),
            math.floor(self.y),
            math.floor(self.z)
        )

    def __ceil__(self) -> Vector3D:
        """Get the ceiling of a vector.

        Returns:
            Vector3D: resulting vector.
        """
        return Vector3D(
            math.ceil(self.x),
            math.ceil(self.y),
            math.ceil(self.z)
        )

    def __trunc__(self) -> Vector3D:
        """Get the truncated value of a vector.

        Returns:
            Vector3D: resulting vector.
        """
        return Vector3D(
            math.trunc(self.x),
            math.trunc(self.y),
            math.trunc(self.z)
        )

    def __eq__(self, other: Any) -> bool:
        """Check if two vectors are equal.

        Args:
            other (Any): vector to compare to.

        Returns:
            bool: whether the two vectors are equal to another.
        """
        return (
            self.x == other.x
            and self.y == other.y
            and self.z == other.z
        )

    def __ne__(self, other: Any) -> bool:
        """Check if two vectors are not equal.

        Args:
            other (Any): vector to compare to.

        Returns:
            bool: whether the two vectors are not equal to another.
        """
        return (
            self.x != other.x
            or self.y != other.y
            or self.z != other.z
        )

    def __lt__(self, other: Any) -> bool:
        """Check if a vector is less than another one.

        Args:
            other (Any): vector to compare to.

        Returns:
            bool: whether the vector is less than another one.
        """
        raise NotImplementedError("< operation not supported for Vector3D")

    def __le__(self, other: Any) -> bool:
        """Check if a vector is less than or equal to another one.

        Args:
            other (Any): vector to compare to.

        Returns:
            bool: whether the vector is less than or equal to another one.
        """
        raise NotImplementedError("<= operation not supported for Vector3D")

    def __gt__(self, other: Any) -> bool:
        """Check if a vector is greater than another one.

        Args:
            other (Any): vector to compare to.

        Returns:
            bool: whether the vector is greater than another one.
        """
        raise NotImplementedError("> operation not supported for Vector3D")

    def __ge__(self, other: Any) -> bool:
        """Check if a vector is greater than or equal to another one.

        Args:
            other (Any): vector to compare to.

        Returns:
            bool: whether the vector is greater than or equal to another one.
        """
        raise NotImplementedError(">= operation not supported for Vector3D")

    def __bool__(self) -> bool:
        """Check if a vector is truthy.

        Returns:
            bool: whether the vector is truthy or not.
        """
        return bool(self.x or self.y or self.z)

    def __len__(self) -> int:
        """Get the length of a vector.

        Returns:
            int: length of the vector.
        """
        return 3

    def __iter__(self) -> Iterator:
        """Get an iterator for the vector.

        Returns:
            Iterator: iterator for the vector.
        """
        return iter((self.x, self.y, self.z))

    def __repr__(self) -> str:
        """Get the raw representation of the vector.

        Returns:
            str: raw representation of the vector.
        """
        return f"Vector3D({self.x}, {self.y}, {self.z})"

    def __str__(self) -> str:
        """Get the string representation of the vector.

        Returns:
            str: string representation of the vector.
        """
        return f"({self.x}, {self.y}, {self.z})"

    def __hash__(self) -> int:
        """Get the hash of the vector.

        Returns:
            int: hash of the vector.
        """
        return hash((self.x, self.y, self.z))


class Rotator3D(Vector3D):
    """3D rotation representation class.

    This class represents a rotation in the 3D space. It can also be used to
    represent rotations in a cartesian coordinate system.

    Attributes:
        x (float): X rotation (degrees).
        y (float): Y rotation (degrees).
        z (float): Z rotation (degrees).
    """

    def __init__(
        self,
        x: int | float = 0,
        y: int | float = 0,
        z: int | float = 0
    ) -> None:
        """Initialize a Rotator3D instance.

        Args:
            x (int | float): X rotation (degrees).
            y (int | float): Y rotation (degrees).
            z (int | float): Z rotation (degrees).
        """
        self.x = np.deg2rad(x)
        self.y = np.deg2rad(y)
        self.z = np.deg2rad(z)

    def __repr__(self) -> str:
        """Get the raw representation of the rotator.

        Returns:
            str: raw representation of the rotator.
        """
        return f"Rotator3D({self.x}, {self.y}, {self.z})"

    def __str__(self) -> str:
        """Get the string representation of the rotator.

        Returns:
            str: string representation of the rotator.
        """
        return f"({self.x}, {self.y}, {self.z})"


def distance3D(a: Vector3D, b: Vector3D) -> float:
    """Get the distance between two vectors.

    Args:
        a (Vector3D): first vector.
        b (Vector3D): second vector.

    Returns:
        float: distance between the two vectors.
    """
    if not (isinstance(a, Vector3D) and isinstance(b, Vector3D)):
        raise TypeError(
            "expected type Vector3D for operands but got"
            + f" {type(a).__name__} and {type(b).__name__} instead"
        )

    return ((b.x - a.x)**2 + (b.y - a.y)**2 + (b.z - a.z)**2) ** .5



"""Simulation API module.

This module combines all API resources to compose the Simulation API,
responsible for the execution, update and summary of the simulation.

Author:
    Paulo Sanchez (@erlete)
"""


import json
import os
from time import perf_counter as pc

import matplotlib.pyplot as plt
import numpy as np
from colorama import Fore, Style
from scoretree import Score, ScoreArea, ScoreTree

from ..core.gradient import ColorGradient
from ..core.vector import Rotator3D, Vector3D, distance3D
from ..environment.track import Track
from .drone import DroneAPI
from .statistics import TrackStatistics
from .track import TrackAPI


class SimulationAPI:
    """Simulation API class.

    This class represents a simulation that implements all kinematic variants
    of the simulation elements, such as the drone and the track. It provides
    with several methods that allow the user to get information about the
    simulation's state and control it.

    Attributes:
        tracks (list[TrackAPI]): track list.
        drone (DroneAPI): drone element.
        next_waypoint (Vector3D | None): next waypoint data.
        remaining_waypoints (int): remaining waypoints in the track.
        is_simulation_finished (bool): whether the simulation is finished.
        DT (float): simulation time step in seconds.
        DV (float): simulation speed step in m/s.
        DR (float): simulation rotation step in rad/s.
    """

    DT = 0.1  # [s]
    DV = 7.5  # [m/s]
    DR = np.pi  # [rad/s]

    SUMMARY_FILE_PREFIX = "summary_"
    SUMMARY_DIR = "statistics"

    def __init__(self, tracks: list[Track]) -> None:
        """Initialize a SimulationAPI instance.

        Args:
            tracks (list[Track]): track list.
        """
        self._completed_statistics: list[TrackStatistics] = []
        self._statistics = [
            TrackStatistics(TrackAPI(track), self.DT)
            for track in tracks
        ]
        self.tracks = [TrackAPI(track) for track in tracks]  # Conversion.

    @property
    def tracks(self) -> list[TrackAPI]:
        """Get track list.

        Returns:
            list[TrackAPI]: track list.
        """
        return self._tracks

    @tracks.setter
    def tracks(self, value: list[TrackAPI]) -> None:
        """Set track list.

        Args:
            value (list[TrackAPI]): track list.
        """
        if not isinstance(value, list):
            raise TypeError(
                "expected type list[Track] for"
                + f" {self.__class__.__name__}.tracks but got"
                + f" {type(value).__name__} instead"
            )

        if not value:
            raise ValueError(
                f"{self.__class__.__name__}.tracks cannot be empty"
            )

        for i, track in enumerate(value):
            if not isinstance(track, TrackAPI):
                raise TypeError(
                    "expected type Track for"
                    + f" {self.__class__.__name__}.tracks but got"
                    + f" {type(track).__name__} from item at index {i} instead"
                )

        self._tracks = value

        # Internal attributes reset:
        self._is_simulation_finished = False
        self._current_track = self._tracks.pop(0)
        self._current_statistics = self._statistics.pop(0)
        self._current_timer = 0.0
        self._target_rotation = Rotator3D()
        self._target_speed = 0.0

    @property
    def drone(self) -> DroneAPI:
        """Returns the drone element.

        Returns:
            DroneAPI: drone element.
        """
        return self._current_track.drone

    @property
    def next_waypoint(self) -> Vector3D | None:
        """Returns the next waypoint data.

        Returns:
            Vector3D | None: next waypoint data.
        """
        return self._current_track.next_waypoint

    @property
    def remaining_waypoints(self) -> int:
        """Returns the remaining waypoints in the track.

        Returns:
            int: remaining waypoints in the track.
        """
        return self._current_track.remaining_waypoints

    @property
    def is_simulation_finished(self) -> bool:
        """Returns whether the simulation is finished.

        Returns:
            bool: True if the simulation is finished, False otherwise.
        """
        return self._is_simulation_finished

    def set_drone_target_state(
        self,
        yaw: int | float,
        pitch: int | float,
        speed: int | float
    ) -> None:
        """Set drone target state.

        Args:
            yaw (int | float): target drone yaw in radians.
            pitch (int | float): target drone pitch in radians.
            speed (int | float): target drone speed in m/s.
        """
        if not isinstance(yaw, (int, float)):
            raise TypeError(
                "expected type (int, float) for"
                + f" {self.__class__.__name__}.set_drone_target_state yaw"
                + f" but got {type(yaw).__name__} instead"
            )

        if not isinstance(pitch, (int, float)):
            raise TypeError(
                "expected type (int, float) for"
                + f" {self.__class__.__name__}.set_drone_target_state pitch"
                + f" but got {type(pitch).__name__} instead"
            )

        if not isinstance(speed, (int, float)):
            raise TypeError(
                "expected type int | float for"
                + f" {self.__class__.__name__}.set_drone_target_state speed"
                + f" but got {type(speed).__name__} instead"
            )

        self._target_rotation = Rotator3D(
            np.rad2deg(yaw),
            np.rad2deg(pitch),
            0
        )
        self._target_speed = speed

    def update(
        self,
        plot: bool = True,
        dark_mode: bool = False,
        fullscreen: bool = True
    ) -> None:
        """Update drone state along the current track and plot environment.

        Args:
            plot (bool): whether to plot statistics after each track. Defaults
                to True.
            dark_mode (bool): whether to use dark mode for the plot. Defaults
                to False. Only used if plot is True.
            fullscreen (bool): whether to plot the figure in fullscreen mode.
                Defaults to True. Only used if plot is True.
        """
        self._current_timer += self.DT

        # Simulation endpoint conditions' definition for later use:
        c1 = self._current_timer >= self._current_track.timeout
        c2 = self._current_track.is_track_finished
        c3 = self._current_track.is_drone_stopped

        # On completed track finish condition:
        if c2 and c3:
            self._current_statistics.is_completed = True
            self._current_statistics.distance_to_end = distance3D(
                self._current_track.drone.position,
                self._current_track.track.end
            )

        # On each of the simulation finish conditions:
        if c1 or (c2 and c3):

            # Plot current track statistics:
            if plot:
                self.plot(dark_mode, fullscreen)

            # Save current statistics:
            self._completed_statistics.append(self._current_statistics)

            # Get next track and reset time counter:
            if self._tracks:
                self._current_track = self._tracks.pop(0)
                self._current_statistics = self._statistics.pop(0)
                self._current_timer = 0.0
            else:
                self._is_simulation_finished = True

            return

        # Rotation update:
        self._current_track.drone.rotation = Rotator3D(
            *[
                np.rad2deg(
                    min(curr_rot + self.DR * self.DT, tg_rot)
                    if curr_rot < tg_rot else
                    max(curr_rot - self.DR * self.DT, tg_rot)
                ) for curr_rot, tg_rot in zip(
                    self._current_track.drone.rotation,
                    self._target_rotation
                )
            ]
        )

        # Speed update:
        speed = self._current_track.drone.speed
        self._current_track.drone.speed = (
            min(speed + self.DV * self.DT, self._target_speed)
            if self._target_speed >= speed else
            max(speed - self.DV * self.DT, self._target_speed)
        )

        # Position update:
        rot = self._current_track.drone.rotation
        self._current_track.drone.position += (
            Vector3D(
                speed * self.DT * np.cos(rot.x) * np.cos(rot.y),
                speed * self.DT * np.sin(rot.x) * np.cos(rot.y),
                speed * self.DT * np.sin(rot.y)
            )
        )

        self._current_statistics.add_data(
            position=self._current_track.drone.position,
            rotation=self._current_track.drone.rotation,
            speed=self._current_track.drone.speed
        )

    def plot(self, dark_mode: bool, fullscreen: bool) -> None:
        """Plot simulation environment.

        Args:
            dark_mode (bool): whether to use dark mode for the plot.
            fullscreen (bool): whether to plot the figure in fullscreen mode.
        """
        # Variable definition for later use:
        times = np.arange(0, self._current_track.timeout, self.DT)
        speeds = self._current_statistics.speeds
        rotations = [
            [rot.x for rot in self._current_statistics.rotations],
            [rot.y for rot in self._current_statistics.rotations],
            [rot.z for rot in self._current_statistics.rotations]
        ]
        positions = [
            [pos.x for pos in self._current_statistics.positions],
            [pos.y for pos in self._current_statistics.positions],
            [pos.z for pos in self._current_statistics.positions]
        ]
        gradient = ColorGradient("#dc143c", "#15b01a", len(positions[0]))

        # Figure and axes setup:
        plt.style.use("dark_background" if dark_mode else "fast")
        fig = plt.figure()
        ax1 = fig.add_subplot(121, projection="3d")
        ax2 = fig.add_subplot(422)
        ax3 = fig.add_subplot(424)
        ax4 = fig.add_subplot(426)
        ax5 = fig.add_subplot(428)

        # 2D axes configuration:
        config_2d = {
            "axes": (ax2, ax3, ax4, ax5),
            "data": (speeds, *rotations),
            "labels": (
                "Speed [m/s]",
                "X rotation [rad]",
                "Y rotation [rad]",
                "Z rotation [rad]"
            ),
            "titles": (
                "Speed vs Time",
                "X rotation vs Time",
                "Y rotation vs Time",
                "Z rotation vs Time"
            )
        }

        for ax, data_, title, label in zip(*config_2d.values()):
            ax.plot(times[:len(data_)], data_)
            ax.set_xlim(0, self._current_track.timeout)
            ax.set_title(title)
            ax.set_xlabel("Time [s]")
            ax.set_ylabel(label)
            ax.grid(True)

        ax2.set_ylim(self._current_track.drone.SPEED_RANGE)

        # 3D ax configuration:
        self._current_track._track.plot(ax1)

        for x, y, z, c in zip(*positions, gradient.steps):
            ax1.plot(x, y, z, "D", color=ColorGradient.rgb_to_hex(c), ms=2)

        ax1.plot(*positions, "k--", alpha=.75, lw=.75)

        ax1.set_title("3D Flight visualization")
        ax1.set_xlabel("X [m]")
        ax1.set_ylabel("Y [m]")
        ax1.set_zlabel("Z [m]")  # type: ignore

        # Figure configuration:
        plt.tight_layout()
        plt.get_current_fig_manager().window.state(  # type: ignore
            "zoomed" if fullscreen else "normal"
        )
        plt.show()

    def _compute_score(
        self,
        statistics: TrackStatistics
    ) -> tuple[bool, list[Score]]:
        """Compute track simulation score from statistics.

        Args:
            statistics (TrackStatistics): track statistics.

        Returns:
            tuple[bool, list[Score]]: list containing track completion flag and
                list of scores on each weighted area.
        """
        # Variable definition for later use:
        waypoints = statistics.track.track.waypoints  # Track waypoints.
        positions = statistics.positions  # Drone positions during simulation.
        max_sp = self.drone.SPEED_RANGE[1]  # Max drone speeed.

        # Track Distance (TD):
        min_td = sum([
            distance3D(waypoints[i], waypoints[i+1])
            for i in range(len(waypoints) - 1)
        ])
        max_td = 2 * min_td
        td = sum([
            distance3D(positions[i], positions[i+1])
            for i in range(len(positions) - 1)
        ])

        # Distance To End (DTE):
        max_tte = max_sp / self.DV  # Max time to end.
        min_dte = 0
        max_dte = max_sp * max_tte - .5 * self.DV * max_tte ** 2
        dte = statistics.distance_to_end

        # Track Time (TT):
        min_tt = max_td / max_sp + min_dte
        max_tt = (max_td / self.DV + max_tte) * 2
        tt = len(statistics.positions) * self.DT

        # Pondered score:
        return (
            statistics.is_completed,
            [
                Score(
                    name="Distance to end (DTE)",
                    weight=0.4,
                    score_range=(min_dte, max_dte),
                    value=dte if statistics.is_completed else max_dte,
                    inverse=True
                ),

                Score(
                    name="Track distance (TD)",
                    weight=0.25,
                    score_range=(min_td, max_td),
                    value=td if statistics.is_completed else max_td,
                    inverse=True
                ),
                Score(
                    name="Track Time (TT)",
                    weight=0.35,
                    score_range=(min_tt, max_tt),
                    value=tt if statistics.is_completed else max_tt,
                    inverse=True
                )
            ]
        )

    def summary(self) -> None:
        """Print a summary of the simulation."""
        # Track weight computation:
        weight_range = range(1, len(self._completed_statistics) + 1)
        track_weights = [
            i / sum(weight_range)
            for i in weight_range
        ]

        # Score tree generation:
        st = ScoreTree([ScoreArea("Simulation", 1, [
            Score(f"Track {i + 1} (DNF)", 1, (0, 1), 0)
            if not score[0] else
            ScoreArea(f"Track {i + 1}", weight, score[1])
            for (i, weight), score in zip(
                enumerate(track_weights), [
                    self._compute_score(stat)
                    for stat in self._completed_statistics
                ]
            )
        ])], colorized=True)

        print(
            Fore.BLUE + Style.BRIGHT
            + " Simulation statistics ".center(80, "=")
            + Style.RESET_ALL + "\n"
            + str(st)
            + Fore.BLUE + Style.BRIGHT + "\n"
            + f" Total score: {st.score * 100:.2f}% ".center(80, "=")
            + Style.RESET_ALL
        )



"""Track API module.

Author:
    Paulo Sanchez (@erlete)
"""


from ..core.vector import Rotator3D, Vector3D, distance3D
from ..environment.track import Track
from .drone import DroneAPI


class TrackAPI:
    """Track API class.

    This class represents a kinematic version of a track, implementing the
    geometry of the track, adding a drone to it and providing information about
    next waypoint, remaining ones, whether the track is finished and whether
    the drone is stopped.

    Attributes:
        track (Track): track.
        drone (DroneAPI): drone.
        waypoints (list[Vector3D]): track waypoints.
        next_waypoint (Vector3D | None): next waypoint data.
        remaining_waypoints (int): remaining waypoints in the track.
        is_track_finished (bool): whether the track is finished.
        is_drone_stopped (bool): whether the drone is stopped.
        timeout (float): track timeout.
        REACHED_THRESHOLD (float): reached waypoint threshold in m.
        MIN_TIMEOUT_SPEED (float): minimum timeout speed in m/s.
    """

    REACHED_THRESHOLD = 1  # [m]
    MIN_TIMEOUT_SPEED = DroneAPI.SPEED_RANGE[1] / 4  # [m/s]

    def __init__(self, track: Track) -> None:
        """Initialize a TrackAPI instance.

        Args:
            track (Track): track.
        """
        self.track = track

        # Internal attributes:
        self._drone = DroneAPI(track.start, Rotator3D())

    @property
    def track(self) -> Track:
        """Get track.

        Returns:
            Track: track.
        """
        return self._track

    @track.setter
    def track(self, value: Track) -> None:
        """Set track.

        Args:
            value (Track): track.
        """
        if not isinstance(value, Track):
            raise TypeError(
                "expected type Track for"
                + f" {self.__class__.__name__}.track but got"
                + f" {type(value).__name__} instead"
            )

        # Internal attributes reset:
        self._waypoints = [*[ring.position for ring in value.rings], value.end]
        self._next_waypoint: Vector3D | None = self._waypoints.pop(0)
        self._is_track_finished = self._is_drone_stopped = False

        self._track = value

    @property
    def drone(self) -> DroneAPI:
        """Get drone.

        Returns:
            DroneAPI: drone.
        """
        return self._drone

    @property
    def waypoints(self) -> list[Vector3D]:
        """Get track waypoints.

        Returns:
            list[Vector3D]: track waypoints.
        """
        return self._waypoints

    @property
    def next_waypoint(self) -> Vector3D | None:
        """Update and return next waypoint data.

        Returns:
            Vector3D | None: next waypoint data or None if the track is
                finished.
        """
        self._eval_reached_waypoint()
        return self._next_waypoint

    @property
    def remaining_waypoints(self) -> int:
        """Get remaining waypoints in the track (including current one).

        Returns:
            int: remaining waypoints in the track (including current one).
        """
        # Adds 1 if the track is not finished to compensate for start waypoint:
        return len(self._waypoints) + (not self._is_track_finished)

    @property
    def is_track_finished(self) -> bool:
        """Check whether the track is finished.

        Returns:
            bool: True if the track is finished, False otherwise.
        """
        return self._is_track_finished

    @property
    def is_drone_stopped(self) -> bool:
        """Check whether the drone is stopped.

        Returns:
            bool: True if the drone is stopped, False otherwise.
        """
        return self._is_track_finished and self._drone.speed == 0

    @property
    def timeout(self) -> float:
        """Get track timeout.

        Track timeout is computed as double the distance between each waypoint
        divided by the minimum expected drone speed.

        Returns:
            float: track timeout.
        """
        return sum(
            distance3D(*self._track.waypoints[i - 1: i + 1])
            for i in range(1, len(self._track.waypoints))
        ) * 2 / self.MIN_TIMEOUT_SPEED

    def _eval_reached_waypoint(self) -> None:
        """Evaluate whether the drone has reached the next waypoint.

        This method is responsible for updating the next waypoint data if the
        drone has reached the current one. It also updates the finished track
        flag if the end of the last waypoint has been reached.
        """
        # Prevent evaluation if the track is finished:
        if self._is_track_finished:
            return

        distance = distance3D(
            self._drone.position,
            self._next_waypoint  # type: ignore
        )

        if distance <= self.REACHED_THRESHOLD:
            if self._waypoints:  # If there are any waypoints left:
                self._next_waypoint = self._waypoints.pop(0)
            else:
                self._next_waypoint = None
                self._is_track_finished = True



"""Track file reader module.

Author:
    Paulo Sanchez (@erlete)
"""


import json

from ..core.vector import Rotator3D, Vector3D
from ..environment.track import Track
from ..geometry.ring import Ring


class TrackSequenceReader:
    """Track sequence reader class.

    This class is used to read a track sequence file and return a list of
    tracks.

    Attributes:
        path (str): track sequence file path.
        track_sequence (list[Track]): track sequence.
    """

    def __init__(self, path: str) -> None:
        """Initialize a TrackSequenceReader instance.

        Args:
            path (str): track sequence file path.
        """
        self.path = path
        self._track_sequence = self.read()

    @property
    def path(self) -> str:
        """Get track sequence file path.

        Returns:
            str: track sequence file path.
        """
        return self._path

    @path.setter
    def path(self, value: str) -> None:
        """Set track sequence file path.

        Args:
            value (str): track sequence file path.
        """
        if not isinstance(value, str):
            raise TypeError(
                "expected type str for"
                + f" {self.__class__.__name__}.path but got"
                + f" {type(value).__name__} instead"
            )

        self._path = value

    @property
    def track_sequence(self) -> list[Track]:
        """Get track sequence.

        Returns:
            list[Track]: track sequence.
        """
        return self._track_sequence

    def read(self) -> list[Track]:
        """Read track sequence file.

        Returns:
            list[Track]: track sequence.
        """
        with open(self.path, mode="r", encoding="utf-8") as fp:
            data = json.load(fp)

        return [
            Track(
                Vector3D(*value["start"]),
                Vector3D(*value["end"]),
                [
                    Ring(
                        Vector3D(*ring["position"]),
                        Rotator3D(*ring["rotation"]),
                    )
                    for ring in value["rings"]
                ]
            ) for value in data.values()
        ]



"""Ring representation module.

Author:
    Paulo Sanchez (@erlete)
"""


import numpy as np

from ..core.vector import Rotator3D, Vector3D


class Ring:
    """Ring representation class.

    Attributes:
        position (Vector3D): ring position.
        rotation (Rotator3D): ring rotation.
        scale (Vector3D): ring scale.
        tube_radius (float): ring tube radius.
        hole_radius (float): ring hole radius.
        complexity (int): ring geometry complexity.
        surface (tuple[np.ndarray, np.ndarray, np.ndarray]): ring geometry
            surface.
    """

    def __init__(
        self,
        position: Vector3D = Vector3D(),
        rotation: Rotator3D = Rotator3D(),
        scale: Vector3D = Vector3D(1, 1, 1),
        tube_radius: int | float = 1,
        hole_radius: int | float = 5,
        complexity: int = 100
    ) -> None:
        """Initialize a Ring instance.

        Args:
            position (Vector3D, optional): ring position. Defaults to
                Vector3D(0, 0, 0).
            rotation (Rotator3D, optional): ring rotation. Defaults to
                Rotator3D(0, 0, 0).
            scale (Vector3D, optional): ring scale. Defaults to
                Vector3D(1, 1, 1).
            tube_radius (int | float, optional): ring tube radius.
                Defaults to 1.
            hole_radius (int | float, optional): ring hole radius.
                Defaults to 5.
            complexity (int, optional): ring geometry complexity. Defaults to
                100.
        """
        self.position = position
        self.rotation = rotation
        self.scale = scale
        self.tube_radius = tube_radius
        self.hole_radius = hole_radius
        self.complexity = complexity

        self._compute_geometry()

    @property
    def position(self) -> Vector3D:
        """Get ring position.

        Returns:
            Vector3D: ring position.
        """
        return self._position

    @position.setter
    def position(self, value: Vector3D) -> None:
        """Set ring position.

        Args:
            value (Vector3D): ring position.
        """
        if not isinstance(value, Vector3D):
            raise TypeError(
                "expected type Vector3D for"
                + f" {self.__class__.__name__}.position but got"
                + f" {type(value).__name__} instead"
            )

        self._position = value

    @property
    def rotation(self) -> Rotator3D:
        """Get ring rotation.

        Returns:
            Rotator3D: ring rotation.
        """
        return self._rotation

    @rotation.setter
    def rotation(self, value: Rotator3D) -> None:
        """Set ring rotation.

        Args:
            value (Rotator3D): ring rotation.
        """
        if not isinstance(value, Rotator3D):
            raise TypeError(
                "expected type Rotator3D for"
                + f" {self.__class__.__name__}.rotation but got"
                + f" {type(value).__name__} instead"
            )

        self._rotation = value

    @property
    def scale(self) -> Vector3D:
        """Get ring scale.

        Returns:
            Vector3D: ring scale.
        """
        return self._scale

    @scale.setter
    def scale(self, value: Vector3D) -> None:
        """Set ring scale.

        Args:
            value (Vector3D): ring scale.
        """
        if not isinstance(value, Vector3D):
            raise TypeError(
                "expected type Vector3D for"
                + f" {self.__class__.__name__}.scale but got"
                + f" {type(value).__name__} instead"
            )

        self._scale = value

    @property
    def tube_radius(self) -> float:
        """Get ring tube radius.

        Returns:
            float: ring tube radius.
        """
        return self._tube_radius

    @tube_radius.setter
    def tube_radius(self, value: int | float) -> None:
        """Set ring tube radius.

        Args:
            value (int | float): ring tube radius.
        """
        if not isinstance(value, (int, float)):
            raise TypeError(
                "expected type int | float for"
                + f" {self.__class__.__name__}.tube_radius but got"
                + f" {type(value).__name__} instead"
            )

        self._tube_radius = float(value)

    @property
    def hole_radius(self) -> float:
        """Get ring hole radius.

        Returns:
            float: ring hole radius.
        """
        return self._hole_radius

    @hole_radius.setter
    def hole_radius(self, value: int | float) -> None:
        """Set ring hole radius.

        Args:
            value (int | float): ring hole radius.
        """
        if not isinstance(value, (int, float)):
            raise TypeError(
                "expected type int | float for"
                + f" {self.__class__.__name__}.hole_radius but got"
                + f" {type(value).__name__} instead"
            )

        self._hole_radius = float(value)

    @property
    def complexity(self) -> int:
        """Get ring geometry complexity.

        Returns:
            int: ring geometry complexity.
        """
        return self._complexity

    @complexity.setter
    def complexity(self, value: int | float) -> None:
        """Set ring geometry complexity.

        Args:
            value (int | float): ring geometry complexity.
        """
        if not isinstance(value, (int, float)):
            raise TypeError(
                "expected type int | float for"
                + f" {self.__class__.__name__}.complexity but got"
                + f" {type(value).__name__} instead"
            )

        self._complexity = int(value)

    @property
    def surface(self) -> np.ndarray:
        """Get ring geometry surface.

        Returns:
            np.ndarray: ring geometry surface.
        """
        return self._surface

    def _compute_geometry(self) -> None:
        """Compute ring geometry.

        This method builds the ring geometry, scales it, rotates it and
        translates it to the correct position.
        """
        theta, phi = np.meshgrid(
            np.linspace(0, 2 * np.pi, self._complexity),  # type: ignore
            np.linspace(0, 2 * np.pi, self._complexity)  # type: ignore
        )

        # X, Y, Z matrix generation:
        matrix = np.array([
            (
                self._hole_radius + self._tube_radius * np.cos(theta)
            ) * np.cos(phi), (
                self._hole_radius + self._tube_radius * np.cos(theta)
            ) * np.sin(phi),
            self._tube_radius * np.sin(theta)
        ])

        # Surface rotation:
        azimuth, elevation, yaw = self._rotation
        rotation_matrix = np.array([
            [
                np.cos(azimuth) * np.cos(elevation),
                (
                    np.cos(azimuth) * np.sin(elevation) * np.sin(yaw)
                    - np.sin(azimuth) * np.cos(yaw)
                ),
                np.cos(azimuth) * np.sin(elevation) * np.cos(yaw)
                + np.sin(azimuth) * np.sin(yaw)
            ], [
                np.sin(azimuth) * np.cos(elevation),
                (
                    np.sin(azimuth) * np.sin(elevation) * np.sin(yaw)
                    + np.cos(azimuth) * np.cos(yaw)
                ),
                np.sin(azimuth) * np.sin(elevation) * np.cos(yaw)
                - np.cos(azimuth) * np.sin(yaw)
            ], [
                -np.sin(elevation),
                np.cos(elevation) * np.sin(yaw),
                np.cos(elevation) * np.cos(yaw)
            ]
        ])
        matrix = np.dot(
            rotation_matrix,
            matrix.reshape(3, -1)
        ).reshape(3, self._complexity, self._complexity)

        # Surface scaling and translation:
        matrix[:, :, :] *= np.array([*self.scale]).reshape(3, 1, 1)
        matrix[:, :, :] += np.array([*self.position]).reshape(3, 1, 1)

        self._surface = matrix

    def plot(self, ax, **kwargs) -> None:
        """Plot ring.

        Args:
            ax (Axes3D): ax to plot on.
        """
        kwargs_ = {
            "rstride": 5,
            "cstride": 5,
            "color": 'k',
            "edgecolors": 'w'
        }
        kwargs_.update(kwargs)

        ax.plot_surface(
            *self._surface,
            **kwargs_
        )

    def __repr__(self) -> str:
        """Get short ring representation.

        Returns:
            str: short ring representation.
        """
        return f"<Ring at {self._position}>"

    def __str__(self) -> str:
        """Get long ring representation.

        Returns:
            str: long ring representation.
        """
        return f"""Ring(
    position={self._position},
    rotation={self._rotation},
    scale={self._scale},
    tube_radius={self._tube_radius},
    hole_radius={self._hole_radius},
    complexity={self._complexity}
)"""



"""Drone representation module.

Author:
    Paulo Sanchez (@erlete)
"""


import numpy as np

from ..core.vector import Rotator3D, Vector3D


class Drone:
    """Drone representation class.

    Attributes:
        position (Vector3D): drone position.
        rotation (Rotator3D): drone rotation.
    """

    STRUCTURE = np.array([
        [0, 0, 0],
        [0, 0, 1],
        [0, 1, 0],
        [0, 1, 1],
        [1, 0, 0],
        [1, 0, 1],
        [1, 1, 0],
        [1, 1, 1]
    ])

    SURFACE = np.array([
        [0, 1, 3, 2],
        [4, 5, 7, 6],
        [0, 1, 5, 4],
        [2, 3, 7, 6],
        [0, 2, 6, 4],
        [1, 3, 7, 5]
    ])

    def __init__(self, position: Vector3D, rotation: Rotator3D) -> None:
        """Initialize a Drone instance.

        Args:
            position (Vector3D): drone position.
            rotation (Rotator3D): drone rotation.
        """
        self.position = position
        self.rotation = rotation

    @property
    def position(self) -> Vector3D:
        """Get drone position.

        Returns:
            Vector3D: drone position.
        """
        return self._position

    @position.setter
    def position(self, value: Vector3D) -> None:
        """Set drone position.

        Args:
            value (Vector3D): drone position.
        """
        if not isinstance(value, Vector3D):
            raise TypeError(
                "expected type Vector3D for"
                + f" {self.__class__.__name__}.position but got"
                + f" {type(value).__name__} instead"
            )

        self._position = value

    @property
    def rotation(self) -> Rotator3D:
        """Get drone rotation.

        Returns:
            Rotator3D: drone rotation.
        """
        return self._rotation

    @rotation.setter
    def rotation(self, value: Rotator3D) -> None:
        """Set drone rotation.

        Args:
            value (Rotator3D): drone rotation.
        """
        if not isinstance(value, Rotator3D):
            raise TypeError(
                "expected type Rotator3D for"
                + f" {self.__class__.__name__}.rotation but got"
                + f" {type(value).__name__} instead"
            )

        self._rotation = value

    def plot(self, ax) -> None:
        """Plot drone.

        Args:
            ax (Axes3D): ax to plot drone on.
        """
        print("Warning: Drone.plot() has not been implemented yet.")
        return None

    def __repr__(self) -> str:
        """Get short drone representation.

        Returns:
            str: short drone representation.
        """
        return f"<Drone at {self._position}>"

    def __str__(self) -> str:
        """Get long drone representation.

        Returns:
            str: long drone representation.
        """
        return f"""Drone(
    position={self._position},
    rotation={self._rotation}
)"""



"""Main simulation module.

This module contains all simulation code. It imports the Simulation API, which
allows the user to control a Drone along a set of designed tracks. The API
automatically handles track switching, so the information it provides is always
related to the current track.

Author:
    Paulo Sanchez (@erlete)
"""

import os

from sdc.api.simulation import SimulationAPI
from sdc.core.vector import Rotator3D, Vector3D, distance3D
from sdc.environment.reader import TrackSequenceReader

# Simulation API definition:
sim = SimulationAPI(
    TrackSequenceReader(
        os.path.join(os.path.dirname(__file__), "tracks.json")
    ).track_sequence
)

# Simulation mainloop:
while not sim.is_simulation_finished:

    """Drone element.

    This element represents the dynamic part of the simulation. It provides
    with several methods that allow the user to get information about the
    drone's state and control it.

    Attributes:
        position (Vector3D): current drone position.
        rotation (Rotator3D): current drone rotation.
        speed (float): current drone speed.
    """
    drone = sim.drone

    """Next waypoint data.

    The next_waypoint provides information about the location (x, y, z) of the
    next waypoint of the track. It can either be a Vector3D or None if the end
    of the track has been reached. On the other hand, the `remaining` element
    provides information on how many waypoints are left in the track. If it
    reaches 0, the track is finished.
    """
    next_waypoint, remaining = sim.next_waypoint, sim.remaining_waypoints

    #########################################################
    # TODO: implement path planning and control logic here. #
    #########################################################

    # Update drone state:
    sim.set_drone_target_state(
        yaw=0,  # TODO: implement value change.
        pitch=0,  # TODO: implement value change.
        speed=0  # TODO: implement value change.
    )

    # Update simulation state:
    sim.update(plot=True, dark_mode=False, fullscreen=False)

# Print simulation statistics summary:
sim.summary()
